
/*******************************************************************************
DISCLAIMER

THIS SOFTWARE IS PROVIDED "AS IS" AND BFG ENGINEERING MAKES NO WARRANTIES
REGARDING THIS SOFTWARE, WHETHER EXPRESS, IMPLIED OR STATUTORY, INCLUDING
BUT NOT LIMITED TO WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE AND NON-INFRINGEMENT. ALL SUCH WARRANTIES ARE EXPRESSLY DISCLAIMED.

WE (BFG ENGINEERING) DO NOT WARRANT THAT THIS SOFTWARE IS FREE FROM CLAIMS
BY A THIRD PARTY OF COPYRIGHT, PATENT, TRADEMARK, TRADE SECRET OR ANY OTHER
INTELLECTUAL PROPERTY INFRINGIMENT.

TO THE MAXIMUM EXTENT PERMITTED NOT PROHIBITED BY LAW, BFG ENGINEERING
SHALL NOT BE LIABLE FOR ANY DIRECT, INDIRECT, SPECIAL, INCIDENTAL OR
CONSEQUENTIAL DAMAGES FOR ANY REASON RELATED TO THIS SOFTWARE, EVEN IF BFG
ENGINEERING HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
PARTICULARLY, UNDER NO CIRCUMSTANCES ARE WE LIABLE FOR ANY OF THE FOLLOWING:
- THIRD PARTY CLAIMS AGAINST YOU FOR LOSSES OR DAMAGES;
- LOSS OF, OR DAMAGE TO YOUR RECORDS OR DATA; OR
- ECONOMIC CONSEQUENTIAL DAMAGES (INCLUDING LOST PROFITS OR SAVINGS) OR
  INCIDENTAL DAMAGES, EVEN IF WE ARE INFORMED OF THEIR POSSIBILITY.

BFG ENGINEERING DOES NOT WARRANT UNINTERRUPTED OR ERROR_FREE OPERATION OF
THIS SOFTWARE.
BFG ENGINEERING HAS NO OBLIGATION TO PROVIDE SERVICE, DEFECT CORRECTION OR
ANY MANTEINANCE OF THIS SOFTWARE; OR TO SUPPLY ANY UPDATES OR ENHANCEMENTS
EVEN IF SUCH ARE, OR LATER BECOME, AVAILABLE.

IF YOU DOWNLOAD OR USE THIS SOFTWARE YOU AGREE TO THESE TERMS, AND DO IT AT
YOUR OWN RISK.
*******************************************************************************/

/*******************************************************************************
File:		gpt_pwm.c
Project:	S5D9_MotorControl
Revision:	1.0
Date:		01-05-17 (DD-MM-YYYY)
Author:		Andrea Ghirardello
Contents:	GPT functions used to implement three-phases pwm
*******************************************************************************/

#define _GPT_PWM_C	// insert define BEFORE including header files


/***********************************************************************************************************************
 * Includes
 **********************************************************************************************************************/

#include "mc_timer_api.h"
#include "gpt_pwm.h"
#include "gpt_pwm_private_api.h"


/***********************************************************************************************************************
 * Macro definitions
 **********************************************************************************************************************/

/** Maximum number of clock counts in 32 bit timer */
#define GPT_MAX_CLOCK_COUNTS_32 (0xFFFFFFFFULL)

/** Maximum number of clock counts in 16 bit timer */
#define GPT_MAX_CLOCK_COUNTS_16 (0xFFFFUL)

/** Macro for error logger. */
#ifndef GPT_ERROR_RETURN
/*LDRA_INSPECTED 77 S This macro does not work when surrounded by parentheses. */
#define GPT_ERROR_RETURN(a, err) SSP_ERROR_RETURN((a), (err), &g_module_name[0], &g_gpt_version)
#endif

/** Variant data mask to determine if GPT is 16-bit (0) or 32-bit (1). */
#define GPT_VARIANT_TIMER_SIZE_MASK  (0x4U)

/** Variant data bit to determine if GPT is base (0) or matsu (1). */
#define GPT_VARIANT_BASE_MATSU_BIT   (3)


/***********************************************************************************************************************
 * Typedef definitions
 **********************************************************************************************************************/

/***********************************************************************************************************************
 * Private function prototypes
 **********************************************************************************************************************/

static void	gpt_pwm_hardware_init		(	mc_timer_instance_ctrl_t	* const p_ctrl,
											mc_timer_cfg_t const		* const p_cfg);
static ssp_err_t gpt_pwm_common_open	(	mc_timer_instance_ctrl_t	* const p_ctrl,
											ssp_feature_t const			* const p_ssp_feature,
											uint16_t					* const p_variant);


/***********************************************************************************************************************
 * Private global variables
 **********************************************************************************************************************/

/** Name of module used by error logger macro */
#if BSP_CFG_ERROR_LOG != 0
static const char g_module_name[] = "gpt_pwm";
#endif

#if defined(__GNUC__)
/* This structure is affected by warnings from a GCC compiler bug.  This pragma suppresses the warnings in this 
 * structure only. */
/*LDRA_INSPECTED 69 S */
#pragma GCC diagnostic ignored "-Wmissing-field-initializers"
#endif
/** Version data structure used by error logger macro. */
static const ssp_version_t g_gpt_pwm_version =
{
    .api_version_minor  = MC_TIMER_API_VERSION_MINOR,
    .api_version_major  = MC_TIMER_API_VERSION_MAJOR,
    .code_version_major = GPT_PWM_CODE_VERSION_MAJOR,
    .code_version_minor = GPT_PWM_CODE_VERSION_MINOR
};
#if defined(__GNUC__)
/* Restore warning settings for 'missing-field-initializers' to as specified on command line. */
/*LDRA_INSPECTED 69 S */
#pragma GCC diagnostic pop
#endif


/***********************************************************************************************************************
 * Global Variables
 **********************************************************************************************************************/

/** GPT Implementation of motor control timer driver */
const mc_timer_api_t g_pwm_timer_on_gpt =
{
    .open            = GPT_PWM_TimerOpen,
    .stop            = GPT_PWM_Stop,
    .start           = GPT_PWM_Start,
    .enableOut       = GPT_PWM_EnableOut,
    .disableOut      = GPT_PWM_DisableOut,
    .outEnabled      = GPT_PWM_OutEnabled,
    .dutyCyclesSet   = GPT_PWM_DutyCyclesSet,
    .dutyCyclesGet   = GPT_PWM_DutyCyclesGet,
    .close           = GPT_PWM_Close,
    .versionGet      = GPT_PWM_VersionGet
};


/***********************************************************************************************************************
 * Functions
 **********************************************************************************************************************/

ssp_err_t GPT_PWM_EnableOut (mc_timer_ctrl_t * const p_api_ctrl)
{
	// enables all outputs
	gtior_t * p_reg;
	p_reg = HW_GPT_GTIOCRegLookup(p_api_ctrl->channels[0].p_reg, GPT_GTIOCB);
	HW_GPT_GTIOCPinOutputEnable(p_reg, 1u);
	p_reg = HW_GPT_GTIOCRegLookup(p_api_ctrl->channels[1].p_reg, GPT_GTIOCB);
	HW_GPT_GTIOCPinOutputEnable(p_reg, 1u);
	p_reg = HW_GPT_GTIOCRegLookup(p_api_ctrl->channels[2].p_reg, GPT_GTIOCB);
	HW_GPT_GTIOCPinOutputEnable(p_reg, 1u);
	p_reg = HW_GPT_GTIOCRegLookup(p_api_ctrl->channels[0].p_reg, GPT_GTIOCA);
	HW_GPT_GTIOCPinOutputEnable(p_reg, 1u);
	p_reg = HW_GPT_GTIOCRegLookup(p_api_ctrl->channels[1].p_reg, GPT_GTIOCA);
	HW_GPT_GTIOCPinOutputEnable(p_reg, 1u);
	p_reg = HW_GPT_GTIOCRegLookup(p_api_ctrl->channels[2].p_reg, GPT_GTIOCA);
	HW_GPT_GTIOCPinOutputEnable(p_reg, 1u);

	return SSP_SUCCESS;
} // End of function GPT_PWM_EnableOut

ssp_err_t GPT_PWM_DisableOut (mc_timer_ctrl_t * const p_api_ctrl)
{
	// disables all outputs
	gtior_t * p_reg;
	p_reg = HW_GPT_GTIOCRegLookup(p_api_ctrl->channels[0].p_reg, GPT_GTIOCA);
	HW_GPT_GTIOCPinOutputDisable(p_reg);	// HW_GPT_GTIOCPinOutputEnable(p_reg, 0);
	p_reg = HW_GPT_GTIOCRegLookup(p_api_ctrl->channels[1].p_reg, GPT_GTIOCA);
	HW_GPT_GTIOCPinOutputDisable(p_reg);	// HW_GPT_GTIOCPinOutputEnable(p_reg, 0);
	p_reg = HW_GPT_GTIOCRegLookup(p_api_ctrl->channels[2].p_reg, GPT_GTIOCA);
	HW_GPT_GTIOCPinOutputDisable(p_reg);	// HW_GPT_GTIOCPinOutputEnable(p_reg, 0);
	p_reg = HW_GPT_GTIOCRegLookup(p_api_ctrl->channels[0].p_reg, GPT_GTIOCB);
	HW_GPT_GTIOCPinOutputDisable(p_reg);	// HW_GPT_GTIOCPinOutputEnable(p_reg, 0);
	p_reg = HW_GPT_GTIOCRegLookup(p_api_ctrl->channels[1].p_reg, GPT_GTIOCB);
	HW_GPT_GTIOCPinOutputDisable(p_reg);	// HW_GPT_GTIOCPinOutputEnable(p_reg, 0);
	p_reg = HW_GPT_GTIOCRegLookup(p_api_ctrl->channels[2].p_reg, GPT_GTIOCB);
	HW_GPT_GTIOCPinOutputDisable(p_reg);	// HW_GPT_GTIOCPinOutputEnable(p_reg, 0);

	return SSP_SUCCESS;
} // End of function GPT_PWM_DisableOut

ssp_err_t GPT_PWM_OutEnabled (mc_timer_ctrl_t * const p_api_ctrl, bool * enabled)
{
    // only checks first channel A output
    gtior_t * p_reg = HW_GPT_GTIOCRegLookup(p_api_ctrl->channels[0].p_reg, GPT_GTIOCA);
    *enabled = HW_GPT_PWM_GTIOCPinOutputEnabled(p_reg);

    return SSP_SUCCESS;
} // End of function GPT_PWM_OutEnabled

ssp_err_t GPT_PWM_TimerOpen (	mc_timer_ctrl_t			* const p_api_ctrl,
								mc_timer_cfg_t const	* const p_cfg)
{
/*
#if GPT_CFG_PARAM_CHECKING_ENABLE
    SSP_ASSERT(NULL != p_api_ctrl);
    SSP_ASSERT(NULL != p_cfg);
#endif // if GPT_CFG_PARAM_CHECKING_ENABLE
*/
	uint16_t variant = 0;
	ssp_err_t err = SSP_SUCCESS;
	ssp_feature_t ssp_feature;
	mc_timer_instance_ctrl_t * p_ctrl;

	// Power on GPTs before setting any hardware registers
	HW_GPT_PWM_POE_PowerOn();

	ssp_feature.id = SSP_IP_GPT;
	ssp_feature.unit = 0U;
	for(uint8_t i = 0; i < 3; i++)
	{
		p_ctrl = &(p_api_ctrl->channels[i]);
		ssp_feature.channel = p_cfg->channels[i];
		err = gpt_pwm_common_open(p_ctrl, &ssp_feature, &variant);
		GPT_ERROR_RETURN((SSP_SUCCESS == err), err);

		gpt_pwm_hardware_init(p_ctrl, p_cfg);
	}

	// link the A/D converter call to the first channel
	p_ctrl = &(p_api_ctrl->channels[0]);
    mc_timer_par_cfg_t * pnt = p_cfg->p_param;
	if(1 < pnt->fre_ratio)
	{
        uint32_t u32a = pnt->fre_ratio - 1;
        u32a <<= 8;
        u32a |= 0x00001080;     // B12:ADTAL=1 B7:IVTC=2
        HW_GPT_PWM_SetSkippingCountSelect(p_ctrl->p_reg, u32a);
    }
    HW_GPT_PWM_SetAdcStartRequest(p_ctrl->p_reg, pnt->con_delay);
	HW_GPT_PWM_ADReqEnable(p_ctrl->p_reg, 1);

    // power on the POE circuit
    // HW_GPT_PWM_POE_PowerOn(); already done

    // configures POE
    //HW_GPT_PWM_SetPOEGGA(0x30000030);
    HW_GPT_PWM_SetPOEGGB(0x30000030);

    return SSP_SUCCESS;
} // End of function GPT_PWM_TimerOpen

ssp_err_t GPT_PWM_Stop (mc_timer_ctrl_t * const p_api_ctrl)
{
/*
#if GPT_CFG_PARAM_CHECKING_ENABLE
    SSP_ASSERT(NULL != p_api_ctrl);
#endif
*/

    // outputs disable
	ssp_err_t dummy = GPT_PWM_DisableOut(p_api_ctrl);
	
    // Stop all timers
	HW_GPT_CounterStartStop(p_api_ctrl->channels[0].p_reg, GPT_STOP);
	HW_GPT_CounterStartStop(p_api_ctrl->channels[1].p_reg, GPT_STOP);
	HW_GPT_CounterStartStop(p_api_ctrl->channels[2].p_reg, GPT_STOP);
	
    return SSP_SUCCESS;
} // End of function GPT_PWM_Stop

ssp_err_t GPT_PWM_Start (mc_timer_ctrl_t * const p_api_ctrl)
{
/*
#if GPT_CFG_PARAM_CHECKING_ENABLE
    SSP_ASSERT(NULL != p_api_ctrl);
#endif
*/

    // get the channel mask
    uint32_t channelMask = 0;
	channelMask |= (1u << p_api_ctrl->channels[0].channel);
	channelMask |= (1u << p_api_ctrl->channels[1].channel);
	channelMask |= (1u << p_api_ctrl->channels[2].channel);

    // Start timer
    HW_GPT_PWM_Start(channelMask);

    return SSP_SUCCESS;
} // End of function GPT_PWM_Start

ssp_err_t GPT_PWM_DutyCyclesSet (   mc_timer_ctrl_t     * const p_api_ctrl,
                                    mc_timer_dutycycles * const duty_cycles)
{
/*
#if GPT_CFG_PARAM_CHECKING_ENABLE
    SSP_ASSERT(NULL != p_api_ctrl);
#endif // if GPT_CFG_PARAM_CHECKING_ENABLE
*/
    HW_GPT_CompareMatchSet(p_api_ctrl->channels[0].p_reg, GPT_GTIOCA, duty_cycles->duties[0]);
    HW_GPT_CompareMatchSet(p_api_ctrl->channels[1].p_reg, GPT_GTIOCA, duty_cycles->duties[1]);
    HW_GPT_CompareMatchSet(p_api_ctrl->channels[2].p_reg, GPT_GTIOCA, duty_cycles->duties[2]);

return SSP_SUCCESS;
} // End of function GPT_PWM_DutyCyclesSet

ssp_err_t GPT_PWM_DutyCyclesGet (   mc_timer_ctrl_t     * const p_api_ctrl,
                                    mc_timer_dutycycles * const duty_cycles)
{
/*
#if GPT_CFG_PARAM_CHECKING_ENABLE
    SSP_ASSERT(NULL != p_api_ctrl);
#endif // if GPT_CFG_PARAM_CHECKING_ENABLE
*/
    HW_GPT_PWM_CompareMatchGet(p_api_ctrl->channels[0].p_reg, &(duty_cycles->duties[0]));
    HW_GPT_PWM_CompareMatchGet(p_api_ctrl->channels[1].p_reg, &(duty_cycles->duties[1]));
    HW_GPT_PWM_CompareMatchGet(p_api_ctrl->channels[2].p_reg, &(duty_cycles->duties[2]));

    return SSP_SUCCESS;
} // End of function GPT_PWM_DutyCyclesGet

ssp_err_t GPT_PWM_Close (mc_timer_ctrl_t * const p_api_ctrl)
{
/*
#if GPT_CFG_PARAM_CHECKING_ENABLE
    SSP_ASSERT(NULL != p_ctrl);
#endif
*/
    for (uint8_t i = 0; i < 3; i++)
    {
        // Cleanup. Disable interrupts, stop counter, and disable output
        HW_GPT_InterruptDisable(p_api_ctrl->channels[i].p_reg);

        gtior_t * p_reg = HW_GPT_GTIOCRegLookup(p_api_ctrl->channels[i].p_reg, GPT_GTIOCA);
        HW_GPT_GTIOCPinOutputDisable(p_reg);	// HW_GPT_GTIOCPinOutputEnable(p_reg, false);
        p_reg = HW_GPT_GTIOCRegLookup(p_api_ctrl->channels[i].p_reg, GPT_GTIOCB);
        HW_GPT_GTIOCPinOutputDisable(p_reg);	// HW_GPT_GTIOCPinOutputEnable(p_reg, false);
        HW_GPT_CounterStartStop(p_api_ctrl->channels[i].p_reg, GPT_STOP);
    }

    return SSP_SUCCESS;
} // End of function GPT_PWM_Close

ssp_err_t GPT_PWM_VersionGet (ssp_version_t * const p_version)
{
/*
#if GPT_CFG_PARAM_CHECKING_ENABLE
    SSP_ASSERT(NULL != p_version);
#endif
*/

    p_version->version_id = g_gpt_pwm_version.version_id;

    return SSP_SUCCESS;
} // End of function GPT_PWM_VersionGet


/*********************************************************************************************************************//**
 * Private Functions
 **********************************************************************************************************************/

static ssp_err_t gpt_pwm_common_open (	mc_timer_instance_ctrl_t	* const p_ctrl,
										ssp_feature_t const			* const p_ssp_feature,
										uint16_t					* const p_variant)
{
/*
#if GPT_CFG_PARAM_CHECKING_ENABLE
    SSP_ASSERT(NULL != p_api_ctrl);
    SSP_ASSERT(NULL != p_cfg);
#endif // if GPT_CFG_PARAM_CHECKING_ENABLE
*/

	fmi_feature_info_t info = {0U};
	ssp_err_t err = SSP_SUCCESS;

	err = g_fmi_on_fmi.productFeatureGet(p_ssp_feature, &info);
	GPT_ERROR_RETURN(SSP_SUCCESS == err, err);
	p_ctrl->p_reg = info.ptr;
	*p_variant = (uint16_t) info.variant_data;
	if (0U == (info.variant_data & GPT_VARIANT_TIMER_SIZE_MASK))
	{
		p_ctrl->variant = TIMER_VARIANT_16_BIT;
	}
	else
	{
		p_ctrl->variant = TIMER_VARIANT_32_BIT;
	}
	p_ctrl->channel = p_ssp_feature->channel;

	return SSP_SUCCESS;
} // End of function gpt_pwm_common_open

static void gpt_pwm_hardware_init (	mc_timer_instance_ctrl_t	* const p_ctrl,
									mc_timer_cfg_t const		* const p_cfg)
{
	GPT_BASE_PTR p_gpt_reg = 			(GPT_BASE_PTR) p_ctrl->p_reg;

	HW_GPT_RegWriteEnable				(p_gpt_reg);

	HW_GPT_InterruptDisable				(p_gpt_reg);
	HW_GPT_CounterStartStop				(p_gpt_reg, GPT_STOP);
	HW_GPT_InterruptClear				(p_gpt_reg);
	HW_GPT_CounterSet					(p_gpt_reg, 0UL);
	HW_GPT_RegisterInit					(p_gpt_reg, p_ctrl->variant);
	HW_GPT_StartSourceSelect			(p_gpt_reg, GPT_TRIGGER_SOURCE_REGISTER_ENABLE);
	HW_GPT_StopSourceSelect				(p_gpt_reg, GPT_TRIGGER_SOURCE_REGISTER_ENABLE);
	HW_GPT_ClearSourceSelect			(p_gpt_reg, GPT_TRIGGER_SOURCE_REGISTER_ENABLE);
	HW_GPT_ModeSet						(p_gpt_reg, GPT_MODE_TW_PWM_2);

    // get configuration pointer
	gpt_pwm_out_cfg_t * gptCfg =		(gpt_pwm_out_cfg_t *) p_cfg->p_extend;

    // configure GTIOA levels
	gtior_t * gtioReg =					HW_GPT_GTIOCRegLookup (p_gpt_reg, GPT_GTIOCA);
	HW_GPT_GTIOCPinOutputEnable			(gtioReg, gptCfg->gtioca.output_enabled);
	HW_GPT_GTIOCPinLevelStoppedSet		(gtioReg, gptCfg->gtioca.stop_level);
	HW_GPT_GTIOCCompareMatchOutputSet	(gtioReg, gptCfg->gtioca.match_level);
	HW_GPT_GTIOCCycleEndOutputSet		(gtioReg, gptCfg->gtioca.cycle_end_level);
	HW_GPT_GTIOCInitialOutputSet		(gtioReg, gptCfg->gtioca.initial_level);
//	(*gtioReg).GTIOR = 0x0003;			// 0x0403

    // configure GTIOB levels
    gtioReg = 							HW_GPT_GTIOCRegLookup (p_gpt_reg, GPT_GTIOCB);
	HW_GPT_GTIOCPinOutputEnable			(gtioReg, gptCfg->gtiocb.output_enabled);
	HW_GPT_GTIOCPinLevelStoppedSet		(gtioReg, gptCfg->gtiocb.stop_level);
	HW_GPT_GTIOCCompareMatchOutputSet	(gtioReg, gptCfg->gtiocb.match_level);
	HW_GPT_GTIOCCycleEndOutputSet		(gtioReg, gptCfg->gtiocb.cycle_end_level);
	HW_GPT_GTIOCInitialOutputSet		(gtioReg, gptCfg->gtiocb.initial_level);
//	(*gtioReg).GTIOR = 0x0013;			// 0x0413

	// configure buffers
	HW_GPT_PWM_Gtccr_SetBufferMode		(p_gpt_reg, GPT_GTIOCA, GPT_PWM_SINGLE_BUFFER);
	HW_GPT_PWM_Gtccr_SetBufferMode		(p_gpt_reg, GPT_GTIOCB, GPT_PWM_SINGLE_BUFFER);
	HW_GPT_PWM_Gtpr_SetBufferMode		(p_gpt_reg, GPT_PWM_NO_BUFFER);
	HW_GPT_PWM_Gtccr_BufferEnable		(p_gpt_reg, true);
	HW_GPT_PWM_Gtpr_BufferEnable		(p_gpt_reg, false);

	p_gpt_reg->GTBER |= 0x0000000C;     // quick hack since the driver does not distinguish yet based on channels (to pick a or b type)
//	p_gpt_reg->GTBER = 0x0005000E;      // 0x00050002 in GPT32

    // driver still not distinguish between GT32, GT32H and GT32E
	p_gpt_reg->GTDTCR = 0x00000101;     //AG since GT32 timers registers don't have bit 8, no problems in writing it

	// get configuration parameters
	mc_timer_par_cfg_t * pnt = 			p_cfg->p_param;
	HW_GPT_TimerCycleSet				(p_gpt_reg, pnt->halfperiod);
	HW_GPT_PWM_SetDeadTime				(p_gpt_reg, pnt->deadtime);
	HW_GPT_InitialCompareMatchSet		(p_gpt_reg, GPT_GTIOCA, (pnt->halfperiod) - 1);
	HW_GPT_InitialCompareMatchSet		(p_gpt_reg, GPT_GTIOCB, (pnt->halfperiod) - 1);
	HW_GPT_CompareMatchSet				(p_gpt_reg, GPT_GTIOCA, (pnt->halfperiod) - 1);
	HW_GPT_CompareMatchSet				(p_gpt_reg, GPT_GTIOCB, (pnt->halfperiod) - 1);

	p_gpt_reg->GTINTAD = 0x20000000;	//AG enable protection of contemporary high levels

} // End of function gpt_pwm_hardware_init

